/*
MODEL.H file for the COMPLEXSYMMBREAK model.
Written by Gary Felder (gfelder@physics.stanford.edu)
Last Modified 3/6/01

This file contains the model specific functions and definitions for the model:
V = lambda (|Phi|^2 - 1/2 v^2)^2
Comments: This model is for investigating symmetry breaking in a complex one-field case. The field Phi is assumed to begin at zero and roll down from there to its minimum at |Phi|^2=1/2 v^2.

The rescalings are set manually and expansion is neglected. So the constants beta, cpl, f0, r, and s have no effect.
The potential in program units is given by V_{pr} = 1/4 (|phi_{pr}|^2 - 1)^2
*/

/*
General comments about the model.h file:
This file contains the following functions - all called externally
modelinfo(FILE *info_) outputs information about the model and model-specific parameters to a file.
modelinitialize() performs any model-specific initialization
potential_energy(int term, float *field_values) calculates the average potential energy density term by term. The variable num_potential_terms (just above this function) specifies how many separate potential terms are used in this model.
dvdf(int fld, int i, int j, int k) calculates the potential term in the equation of motion, dV/dfield, for the field fld at the lattice point (i,j,k)
effective_mass(float mass_sq[], float *field_values) calculates the square masses of the fields and puts them in the array mass_sq. The parameter beginning tells the function to use initial field values - if this parameter is zero then the field quantities will be calculated dynamically.
model_output(int flush, char *ext_) allows each model to include its own specialized output function(s). The parameter flush is set to 1 when infrequent calculations are being performed and 0 otherwise. The string ext_ gives the extension for output filenames.
*/

/* - This section should be copied into parameters.h when using this model
// ---Adjustable parameters for COMPLEXSYMMBREAK model--- //
const float lambda=1.e-1; // Self-coupling of inflaton
// The following options control output functions
const int sstrings=1; const float tstrings=20.; const float string_cutoff=.01; // Find "strings", i.e. points with |phi|^2<string_cutoff. Start when t>=tstrings
const int srmshistograms=1; // Generate histograms of rms values of phi
*/

// Rescaling parameters.
// The program variables ("_pr") are defined as
//   f_pr = rescale_A a^rescale_r f (f=field value)
//   x_pr = rescale_B x (x=distance)
//   dt_pr = rescale_B a^rescale_s dt (t=time)
// The constants beta, cpl, and f0 are used to set the variable rescalings rescale_A, rescale_B, rescale_r, and rescale_s.
// These rescaling constants may be reset independently; their settings in terms of beta, cpl, and f0 are suggestions. See the documentation for more details.
// These rescalings are intrinsic to the model and probably shouldn't be changed for individual runs. Adjustable parameters are stored in "parameters.h"
const float beta=4.; // Exponent of the dominant term in the potential
const float cpl=lambda; // Coefficient of the dominant term in the potential (up to numerical factors - see documentation)
const float f0=1.; // Initial value of phi in Planck units, typically the point at which phi'=0
// By default these are automatically set to A=1/f0, B=sqrt(cpl) f0^(-1+beta/2), R=6/(2+beta), S=3(2-beta)/(2+beta). They may be adjusted to different values, but the relationship S=2R-3 must be maintained for the program equations to remain correct.
const float rescale_A=1.;
const float rescale_B=sqrt(lambda);
const float rescale_r=1.;
// The value of S in terms of R SHOULD NOT be changed.
const float rescale_s=2.*rescale_r-3.;

// Other global variables
// The array model_vars is intended to hold any model-specific, non-constant global variables.
// These variables should be initialized in modelinitialize() below
// Even if you're not using any, num_model_vars should be at least 1 to keep some compilers happy.
const int num_model_vars=1;
// Model specific variables: None are defined for this model.
extern float model_vars[num_model_vars];

// Macros to make the equations more readable: The only allowed value of fld is 0=Phi
#define PHIR FIELD(0)
#define PHII FIELD(1)
#define PHI2 (pw2(PHIR)+pw2(PHII))

// Model specific details about the run to be output to an information file
inline void modelinfo(FILE *info_)
{
  // Name and description of model
  fprintf(info_,"Complex Symmetery Breaking Model\n");
  fprintf(info_,"V = lambda (|Phi|^2 - 1/2 v^2)^2\n\n");

  // Model specific parameter values
  fprintf(info_,"lambda = %e\n",lambda);
}

// Perform any model specific initialization
// This function is called twice, once before initializing the fields (which_call=1) and once after (which_call=2)
inline void modelinitialize(int which_call)
{
  if(which_call==1)
  {
    if(nflds!=2)
    {
      printf("Number of fields for PLAINSYMMBREAK model must be 2. Exiting.\n");
      exit(1);
    }
    if(expansion!=0)
    {
      printf("Expansion is not implemented for PLAINSYMMBREAK model. Rerun with expansion=0.\n");
      exit(1);
    }
  }
}

// The constant num_potential_terms must be defined for use by outside functions
// Terms: term=0: lambda (|Phi|^2 - 1/2 v^2)^2
const int num_potential_terms=1; // Number of terms that are calculated separately in the potential
// Potential energy terms
// See documentation for normalization of these terms.
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline float potential_energy(int term, float *field_values)
{
  DECLARE_INDICES
  float potential=0.;
  
  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    // Loop over grid to calculate potential term
    LOOP
      potential += pw2(PHI2-1.);
        
    // Convert sum to average
    potential /= gridsize;
  }
  else // If field values are given then use them instead
    potential = pw2(pw2(field_values[0])+pw2(field_values[1])-1.);
  
  // Include numerical coefficients
  potential *= .25;
  
  return (potential);
}

// Potential terms in the equations of motion, dV/dfield, evaluated at point (i,j,k)
// See documentation for details on the normalization of these terms
inline float dvdf(int fld, INDEXLIST)
{
  if(fld==0)
    return( (PHI2-1.)*PHIR );
  else
    return( (PHI2-1.)*PHII );
}

// Calculate effective mass squared and put it into the array mass_sq[] (used for initial conditions and power spectra)
// See documentation for normalization of these terms
// When setting initial conditions field values will be supplied in the array field_values. Otherwise this function will calculate them on the lattice.
inline void effective_mass(float mass_sq[], float *field_values)
{
  DECLARE_INDICES
  int fld;
  float fldsqrd[nflds]; // Square value of field
  float correction; // Used to adjust masses by the appropriate power of the scale factor. (See documentation.)
  
  // Loop over fields to find mean-square value
  if(field_values==NULL) // If no values are given calculate averages on the lattice
  {
    for(fld=0;fld<nflds;fld++)
    {
      fldsqrd[fld]=0.;
      LOOP
        fldsqrd[fld]+=pw2(FIELD(fld));
          
      fldsqrd[fld] /= (float)gridsize;
    }
  }
  else // If field values are given then use them instead
  {
    for(fld=0;fld<nflds;fld++)
      fldsqrd[fld]=pw2(field_values[fld]);
  }
  
  mass_sq[0] = 3.*fldsqrd[0] + fldsqrd[1] - 1.;
  mass_sq[1] = 3.*fldsqrd[1] + fldsqrd[0] - 1.;
  
  // Avoid tachyonic mass term
  if(mass_sq[0]<0.)
    mass_sq[0]=0.;
  
  // Put in scale factor correction. This calculation should be the same for all models.
  if(expansion>0) // If there's no expansion don't bother with this.
  {
    correction = pow(a,2.*rescale_s+2.);
    for(fld=0;fld<nflds;fld++)
      mass_sq[fld] *= correction;
  }
}

////////////////////////////////////////////////////////////////////////////////////
// Model-specific output functions - The function model_output is called externally/
////////////////////////////////////////////////////////////////////////////////////

// This function finds all points on the lattice in "strings," i.e. with magnitude(sigma) less than a preset cutoff
inline void find_strings(char *ext_)
{
  static FILE *strings_,*stringtimes_;
  char name_[500]; // Used for filenames
  int i,j,k;
  int numpts=0; // Number of points in strings at a given time

  static int first=1;
  if(first) // Open output files
  {
    sprintf(name_,"strings%s",ext_);
    strings_=fopen(name_,mode_);
    sprintf(name_,"stringtimes%s",ext_);
    stringtimes_=fopen(name_,mode_);
    first=0;
  }

  LOOP
  {
    if(PHI2<string_cutoff) // All values below the cutoff are considered part of a string
    {
      if(NDIMS==1)
        fprintf(strings_,"%d\n",i);
      else if(NDIMS==2)
        fprintf(strings_,"%d %d\n",i,j);
      else if(NDIMS==3)
        fprintf(strings_,"%d %d %d\n",i,j,k);
      numpts++;
    }
  }
  fprintf(stringtimes_,"%f %d\n",t,numpts);
  fflush(strings_);
  fflush(stringtimes_);

  return;
}

// This function outputs histograms of the rms values of all complex fields
// This array rmshistflds lists, pairwise, the real and imaginary parts of each complex field. It must have an even number of entries
const int rmshistflds[]={0,1};
inline void rms_histograms(char *ext_)
{
  static FILE **rmshist_,*rmshisttimes_;
  char name_[500]; // Used for filenames
  int i,j,k,pair; // pair iterates over pairs of fields, each of which consists of f_R and f_I
  int fld1,fld2; // Indices of the fields for a given histogram
  int binnum; // Index of bin for a given field value
  float binfreq[nbins]; // The frequency of field values occurring within each bin
  float bmin,bmax,df; // Minimum and maximum field values for each field and spacing (in field values) between bins
  float rms; // RMS value of the complex field. Using this variable avoids recalculating it several times at each point.
  
  static int first=1,numhists;
  if(first) // Open output files
  {
    numhists = (sizeof(rmshistflds)/sizeof(rmshistflds[0])/2); // Number of pairs in field list
    rmshist_ = new FILE *[numhists]; // Allocate an array of FILE pointers
    for(pair=0;pair<numhists;pair++)
    {
      sprintf(name_,"rmshist%d%s",pair,ext_);
      rmshist_[pair]=fopen(name_,mode_);
    }
    sprintf(name_,"rmshisttimes%s",ext_);
    rmshisttimes_=fopen(name_,mode_);
    first=0;
  }
  
  fprintf(rmshisttimes_,"%f",t); // Output time at which power spectra were recorded
  
  for(pair=0;pair<numhists;pair++)
  {
    i=0;j=0;k=0;
    fld1=2*pair;
    fld2=2*pair+1;
    // Find the minimum and maximum rms values
    bmin=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2)));
    bmax=bmin;
    LOOP
    {
      rms=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2))); // Calculate RMS value
      bmin = (rms<bmin ? rms : bmin);
      bmax = (rms>bmax ? rms : bmax);
    }
        
    // Find the difference (in field value) between successive bins
    df=(bmax-bmin)/(float)(nbins); // bmin will be at the bottom of the first bin and bmax at the top of the last
        
    // Initialize all frequencies to zero
    for(i=0;i<nbins;i++)
      binfreq[i]=0.;
        
    // Iterate over grid to determine bin frequencies
    LOOP
    {
      rms=sqrt(pw2(FIELD(fld1))+pw2(FIELD(fld2))); // Calculate RMS value
      binnum=(int)((rms-bmin)/df); // Find index of bin for each value
      if(rms>=bmax) // The maximal field value is at the top of the highest bin
        binnum=nbins-1;
      if(binnum>=0 && binnum<nbins) // Increment frequency in the appropriate bin
        binfreq[binnum]++;
      else // This should never happen, but it doesn't hurt to have an extra check.
      {
        printf("Error in rmshist routine. Invalid bin number.\n");
        fprintf(rmshist_[pair],"Invalid bin number\n");
      }
    } // End of loop over grid
        
    // Output results
    for(i=0;i<nbins;i++)
      fprintf(rmshist_[pair],"%e\n",binfreq[i]/(float)gridsize); // Output bin frequency normalized so the total equals 1
    fprintf(rmshist_[pair],"\n"); // Stick a blank line between times to make the file more readable
    fflush(rmshist_[pair]);
    fprintf(rmshisttimes_," %e %e",bmin,df); // Output the starting point and stepsize for the bins at each time
  } // End of loop over pairs
  fprintf(rmshisttimes_,"\n");
  fflush(rmshisttimes_);
  
  return;
}

inline void model_output(int flush,char *ext_)
{
  if(sstrings && flush && t>=tstrings)
    find_strings(ext_);
  if(srmshistograms && flush) // Make histograms of rms values
    rms_histograms(ext_);
}

#undef PHIR
#undef PHII
#undef PHI2

